# ************************************************************************#
# Title:                    Sego Router Verbs                             #
# Description:              Constants for HTTP verbs                      #
# Author:                   Kabelo Masemola <kn3rdydad@gmail.com>         #
# Original Date:            10 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from ..Exceptions.HTTPVerbException import UnknownHTTPMethodException
class Verb:
    HTTP_GET = 1
    HTTP_POST = 2
    HTTP_PUT = 3
    HTTP_DELETE = 4

    def validate(self,str_method):
        str_method = str_method.lower().strip()
        if str_method == 'get':
            return self.HTTP_GET
        elif str_method == 'post':
            return self.HTTP_POST
        elif str_method == 'put':
            return self.HTTP_PUT
        elif str_method == 'delete':
            return self.HTTP_DELETE
        else:
            raise UnknownHTTPMethodException()




# ************************************************************************#
# Title:                    MiddlewareHandler API                         #
# Description:              This class is the main interface into  Sego's #
#                           middleware system                             #
# Author:                   Sambe Consulting <development@sambe.co.za>    #
# Original Date:            10 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from .Middleware import Middleware
from .Middleware import Middleware as SegoMiddleware
from ..Exceptions.MiddlewareException import *
from singleton_decorator import singleton
from webob import Request, Response
from ..Routing.Route import *

@singleton
class MiddlewareManager(object):

    def __init__(self):
        self.global_middleware = {}
        self.pre_process_middleware = []
        self.post_process_middleware = []

    def load(self, middleware_list):
        """
        This method is used to load middleware into the system
        :param middleware_list:
        :return:
        """
        overwrite = False
        for middleware in middleware_list:
            if "overwrite" in middleware:
                if middleware["overwrite"] in [True, False]:
                    overwrite = middleware["overwrite"]
                else:
                    raise MiddlewareOverwriteMustBeBoolean()

            if "name" in middleware:
                if isinstance(middleware["name"], str):
                    middleware_name = middleware["name"]
                else:
                    raise MiddlewareNameMustBeString()
            else:
                raise MiddlewareMustHaveName()

            if "middleware" in middleware:
                if issubclass(middleware["middleware"], SegoMiddleware):
                    middlewareObject = middleware["middleware"]
                else:
                    raise MiddlewareMustExtend()
            else:
                raise SegoBaseException("All dictionaries must have middleware class ")

            self.add(middleware_name=middleware_name, middleware=middlewareObject, overwrite=overwrite)

    def add(self, middleware_name: str, middleware: Middleware, overwrite=False):
        """
        This method registers a new middleware to be used in the platform
        :param middleware_name:
        :param middleware:
        :param overwrite:
        :return:
        """

        if overwrite == True:
            self.global_middleware[middleware_name] = middleware
        else:
            if middleware_name not in self.global_middleware:
                self.global_middleware[middleware_name] = middleware
            else:
                raise MiddlewareAlreadyRegisteredException()

    def register(self, stage: int, middleware_names: list):
        main_error = "Middleware %s was not loaded into the global middleware registry, use MiddlewareManager.load(self, middleware_list)"

        if stage == Middleware.PREPROCESS:
            for middleware_name in middleware_names:
                if self.validate(middleware_name):
                    if middleware_name not in self.pre_process_middleware:
                        self.pre_process_middleware.append(middleware_name)
                    else:
                        raise SegoBaseException("Middleware %s is registered more than once"% str(middleware_name))
                else:
                    raise SegoBaseException(main_error%middleware_name)

        elif stage == Middleware.POSTPROCESS:
            for middleware_name in middleware_names:
                if self.validate(middleware_name):
                    if middleware_name not in self.post_process_middleware:
                        self.post_process_middleware.append(middleware_name)
                    else:
                        raise SegoBaseException("Middleware %s is registered more than once"% str(middleware_name))
                else:
                    raise SegoBaseException(main_error%middleware_name)

        else:
            raise SegoBaseException("Use 'Middleware.PREPROCESS or Middleware.POSTPROCESS for the stage argument")

    def get(self, middleware_name):
        """
        This method returns a middleware associated with the 'middleware_name' :parameter
        if the name does not exist a MiddlewareNotFound exception is raised
        :param middleware_name:
        :return: middleware : Sego.Middleware.Middleware
        """
        if middleware_name in self.global_middleware:
            return self.global_middleware[middleware_name]
        else:
            raise MiddlewareNotFound()

    def get_all(self):
        """
        This method returns all registered middleware
        :return: global_middleware : dict
        """
        return self.global_middleware

    def validate(self, middleware_name):
        """
        This method is a lightweight version of get, because it does not return the middleware
        object only a TRUE/FALSE whether it exists or not
        :param middleware_name:
        :return: TRUE/FALSE
        """
        exists = False

        for entry in self.global_middleware:
            if middleware_name.strip().lower() == entry.strip().lower():
                exists = True
                break
        return exists

    def diff(self,middleware_list,excluded_middleware_list):
        list_difference = []
        for middleware in middleware_list:
            if middleware not in excluded_middleware_list:
                list_difference.append(middleware)
        return list_difference


    def process_middleware(self,stage :int ,route,request : Request = None,response : Response = None):
        if stage == Middleware.PREPROCESS:
            route_excluded_middleware = route["excluded_middleware"]["pre_process"]
            route_middleware = route["middleware"]["pre_process"]
            self.run(stage=stage,middleware_list=self.diff(self.pre_process_middleware,route_excluded_middleware),request=request,response=response)
            self.run(stage=stage,middleware_list=self.diff(route_middleware,route_excluded_middleware),request=request,response=response)
        elif stage == Middleware.POSTPROCESS:
            route_excluded_middleware = route["excluded_middleware"]["post_process"]
            route_middleware = route["middleware"]["post_process"]
            self.run(stage=stage,middleware_list=self.diff(self.post_process_middleware,route_excluded_middleware),request=request,response=response)
            self.run(stage=stage,middleware_list=self.diff(route_middleware, route_excluded_middleware),request=request,response=response)
        else:
            raise SegoBaseException("Use 'Middleware.PREPROCESS or Middleware.POSTPROCESS for the stage argument")



    def run(self,stage:int, middleware_list: list, request: Request, response: Response):
        """
        This method instantiates each middleware class in list the runs
        :param middleware_list:
        :param request:
        :param response:
        :return:
        """

        for middleware in middleware_list:

            instance = self.global_middleware[middleware]()
            if stage == Middleware.PREPROCESS:
                print("Pre process middleware")
                instance.process_request(request=request)
            elif stage == Middleware.POSTPROCESS:
                print("Post process middleware")
                instance.process_response(request=request, response=response)



# ************************************************************************#
# Title:                    Sego  API                                     #
# Description:              This class is the main interface into  Sego   #
# Author:                   Sambe Consulting <development@sambe.co.za>    #
# Original Date:            10 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#
from wsgiadapter import WSGIAdapter as RequestsWSGIAdapter
from .Response.Response import Response as SegoResponse
from requests import Session as RequestsSession
from orator import DatabaseManager,Schema,Model
from singleton_decorator import singleton
from .Routing.Router import Router
from whitenoise import WhiteNoise
from confo.Confo import Confo
import confo.Backends as BE
from webob import Request
from .Exceptions import *
from .Middleware import *
import uuid, os
import inspect




@singleton
class Sego:
    def __init__(self, application_name=None):
        self.database = None
        self.router = Router()
        self.configuration_manager = Confo()
        if application_name == None:
            application_name = str(uuid.uuid4())
        self.application_name = application_name
        self.view_environment = None
        self.exception_handlers = {}
        self.app_exception = None
        self.static_file_manager = None
        self.middleware_manager = MiddlewareManager()
        self.middleware = None
        self.route_parameters = None
        self.session_kwargs = None

    def __call__(self, environ, start_response):
        """
        This method defines the callable interface into sego
        :param environ:
        :param start_response:
        :return:
        """
        return self.static_file_manager(environ, start_response)

    def wsgi_app(self, environ, start_response) -> Response:
        """
        This method defines an entry point into the WSGI (Web Server Gateway Interface)
        application
        :param environ:
        :param start_response:
        :return: response : webob.Response
        """

        request = Request(environ)
        self.route_parameters, self.session_kwargs = self.router.find_route(request=request)
        if self.route_parameters is not None:
            self.middleware_manager.process_middleware(stage=Middleware.PREPROCESS,\
                                                       route=self.route_parameters,\
                                                       request=request,\
                                                       response=None)

        response = self.handle_request(request)
        if self.route_parameters is not None:
            self.middleware_manager.process_middleware(stage=Middleware.POSTPROCESS,\
                                                   route=self.route_parameters,\
                                                   request=request,\
                                                   response=response)
        return response(environ, start_response)

    def get_application_name(self) -> str:
        """
        This simple getter returns the application name whether generated or set at Sego's instantiation
        :return: application_name
        """
        return self.application_name

    def route(self, route) -> object:
        """
         This wrapper method is used to define a Flask style routing interface.
         Instead of taking in strings like flask we are taking in Route object
        :param route:
        :return: handler : object
        """

        def wrapper(handler):
            """
            To make sure that the Flask style routes are can be executed by the normal router
            the route is dynamically updated by adding 'SegoBaseController' and the name of the wrapped
            method to the route.The we use the configuration manager to store the object ID for later usage
            :param handler:
            :return:
            """
            try:
                action = str(handler.__name__)
                route.set_action(action=action)
                route.set_controller(controller="SegoBaseController")
                self.configuration_manager.set("dynamic_routes", action, id(handler))
                self.router.add_route(route)
            except Exception as e:
                self.handle_exceptions({}, {}, e)

            return handler

        return wrapper

    def register_routes(self, routes_package):
        """
        This method loads all defined routes into the Sego application
        :param routes_package:
        :return:
        """
        self.router.register_routes(routes_package)

    def get_router_object(self):
        """
        This method returns the router object
        :return: router : Router
        """
        return self.router

    def register_views(self, view):
        """
        This method loads all defined views/templates into the Sego application
        :param view:
        :return:
        """
        self.view_environment = view

    def get_view_environment(self):
        """
        This method returns the loaded view environment
        :return: view_environment : jinja2.Environment
        """
        return self.view_environment

    def register_configurations(self, credentials, backend_type=BE.FILE_BACKEND):
        """
        This method starts the configuration manager, using confo (https://github.com/sambe-consulting/confo)
        It defaults to FILE_BACKEND but ZOOKEEPER_BACKEND can be used too
        :param credentials:
        :param backend_type:
        :return:
        """
        self.configuration_manager.load_backend(credentials=credentials, name=self.application_name,
                                                backend_type=backend_type)
        self.configuration_manager.activate_backend(backend_name=self.application_name)
        self.configuration_manager.create_namespace(self.application_name)
        self.configuration_manager.use_namespace(self.application_name)

    def handle_request(self, request):
        """
        This method handles all requests, by using the Sego Router object
        :param request:
        :return: webob.Response
        """
        response = SegoResponse()
        route_parameters = self.route_parameters
        kwargs = self.session_kwargs


        try:
            if route_parameters is not None:
                handler = self.router.build_handler(route_parameters=route_parameters)
                handler(request, response, **kwargs)
            else:
                raise NotFoundException404("Page not found")
        except Exception as e:
            self.handle_exceptions(request, response, e)

        return response

    def add_exception_handler(self, exception, handler, overwrite=False):
        """
        This method adds custom exception handlers to the Sego system.
        This defines a simple interface to handle and manage exceptions typicall
        to the HTTP request handling
        :param exception:
        :param handler:
        :param overwrite:
        :return:
        """
        if inspect.isclass(exception):
            exception = exception.__name__

        if exception in self.exception_handlers:
            if overwrite:
                self.exception_handlers[exception] = handler
            else:
                raise Exception("Exception already registered, to force the registration, set 'overwrite' to True")
        else:
            self.exception_handlers[exception] = handler

    def register_exception_handlers(self, exception_path):
        """
        This method loads all defined handlers into the current session
        :param exception_path:
        :return:
        """
        self.app_exception = __import__(exception_path)

    def get_exception_handlers(self):
        """
        This method returns loaded exception handlers
        :return: exception_handlers : dict
        """
        return self.exception_handlers

    def handle_exceptions(self, request, response, exception):
        """
        This method handles exceptions as they are thrown through out
        the application code base.This system will match a known exception
        to the appropriate handler if a handler is not found then the error
        with be raised again to be handled by the developer
        :param request:
        :param response:
        :param exception:
        :return:
        """
        exception_name = exception.__class__.__name__

        if exception_name in self.exception_handlers:
            handler = self.exception_handlers[exception_name]
            handler(request, response, exception)
        else:
            raise exception

    def register_static_files(self,static_dir):
        """
        This method registers a directory where static_files (images,css,js,...) are stored,
        this system allows us to serve static files from Sego
        :param static_dir:
        :return:
        """
        self.static_file_manager = WhiteNoise(self.wsgi_app, root=static_dir)

    def register_middleware(self,middleware):
        self.middleware = __import__(middleware)

    def register_database(self):
        config = self.configuration_manager.get('database')
        self.database = DatabaseManager(config)
        self.schema = Schema(self.database)
        Model.set_connection_resolver(self.database)

    def test_session(self, base_url="http://segotestserver"):
        """
        This methid defines a dummy session to be used by pytest
        :param base_url:
        :return: session: RequestSession
        """
        session = RequestsSession()
        session.mount(prefix=base_url, adapter=RequestsWSGIAdapter(self))
        return session



# ************************************************************************#
# Title:                    Sego Response API                             #
# Description:              This class handles response objects           #
# Author:                   Sambe Consulting <development@sambe.co.za>    #
# Original Date:            20 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from webob import Response as WebObResponse
import json
from webob.exc import *

class Response:
    def __init__(self):
        """
        Define response types
        """
        self.json = None
        self.html = None
        self.text = None
        self.redirect = None
        self.content_type = None
        self.body = b''
        self.status_code = 200

    def __call__(self, environ, start_response):
        self.set_body_and_content_type()
        if self.redirect is not None:
            response = WebObResponse(
                body=self.body, content_type=self.content_type, status=self.status_code
            )
        else:
            response = HTTPTemporaryRedirect(
                location = self.redirect
            )
        return response(environ, start_response)

    def set_body_and_content_type(self):
        if self.json is not None:
            self.body = json.dumps(self.json).encode("UTF-8")
            self.content_type = "application/json"

        if self.html is not None:
            self.body = self.html.encode()
            self.content_type = "text/html"

        if self.text is not None:
            self.body = self.text
            self.content_type = "text/plain"




from .SegoBaseException import SegoBaseException


class MiddlewareAlreadyRegisteredException(SegoBaseException):
    """
    This class defines an exception that occurs when the same string is used to
    represent multiple entries of middleware into the Router
    """
    message = "A middleware with this name already exist, use a different name"


class MiddlewareNotFound(SegoBaseException):
    """
    This class defines an exception that occurs when a middleware does not exist
    """
    message = "A middleware with this name does not exist,.."


class MiddlewareNameMustBeString(SegoBaseException):
    """
    This class defines an exception that occurs when a type other than 'str' is sent to middleware
    loader as a name
    """

    message = "The 'middleware_name' argument must be of type  'str' "


class MiddlewareOverwriteMustBeBoolean(SegoBaseException):
    """
    This class defines an exception that occurs when
    """
    message = "The 'overwrite' must be of type  'boolean' "


class MiddlewareMustHaveName(SegoBaseException):
    message = "The middleware must have a 'name' key with value of type 'str' "


class MiddlewareMustExtend(SegoBaseException):
    message = "All middleware must extend Sego.Middleware"






# ************************************************************************#
# Title:                    Sego Route                                    #
# Description:              This class handles definition of routes       #
# Author:                   Kabelo Masemola <kn3rdydad@gmail.com>         #
# Original Date:            10 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from .Verb import Verb
from ..Exceptions import *
from ..Middleware import Middleware,MiddlewareManager

class Route:
    def __init__(self, name: str = None, verb: int = None, controller: str = None, action: str = None,
                 url: str = None):
        self.name = name
        self.verb = verb
        self.controller = controller
        self.action = action
        self.url = url
        self.middleware = {"pre_process": [], "post_process": []}
        self.middleware_manager = MiddlewareManager()
        self.excluded_middleware = {"pre_process":[],"post_process":[]}

    def set_name(self, name) -> object:
        """
        This method sets the name of the route object
        :param name:
        :return: self / Route
        """
        self.name = name
        return self

    def get_name(self) -> str:
        """
        This method returns the name of the route object
        :return: name
        """
        return self.name

    def set_verb(self, verb) -> object:
        """
        This method sets the http verbs used by the route
        :param verb:
        :return: self / Route
        """
        self.verbs = verb
        return self

    def get_verb(self) -> int:
        """
        This method returns the http verbs used by the current route
        :return: verbs
        """
        return self.verb

    def set_controller(self, controller) -> object:
        """
        This method sets the controller which defines the action for this route
        :param controller:
        :return: self / Route
        """
        self.controller = controller
        return self

    def get_controller(self) -> str:
        """
        This method returns the controller for this route
        :return: controller
        """
        return self.controller

    def set_action(self, action) -> object:
        """
        This method sets the action to execute for this route
        :param action:
        :return: self / Route
        """
        self.action = action
        return self

    def get_action(self) -> str:
        """
        This method returns the action for this route
        :return: action
        """
        return self.action

    def set_middleware(self,stage:int ,middleware_list : list) -> object:
       """
        This method sets the list of middleware to be executed before and after the http request
        is handled by the controller
       :param stage:
       :param middleware_list:
       :return: self / Route
       """
       if stage == Middleware.PREPROCESS:
           self.middleware["pre_process"] = self.middleware["pre_process"]+middleware_list
       elif stage == Middleware.POSTPROCESS:
           self.middleware["post_process"] = self.middleware["post_process"]+ middleware_list
       return self

    def get_middleware(self) -> dict:
        """
        This method returns middleware
        :return:  middleware
        """
        return self.middleware

    def unset_middleware(self,stage:int,middleware_list:list) -> object:
        """
        This method allows for unsetting global middleware for the particular route
        :param stage:
        :param middleware_list:
        :return:
        """
        if stage == Middleware.PREPROCESS:
            self.excluded_middleware["pre_process"] = self.excluded_middleware["pre_process"]+middleware_list
        elif stage == Middleware.POSTPROCESS:
            self.excluded_middleware["pre_process"] = self.excluded_middleware["pre_process"]+middleware_list
        return self

    def get_excluded_middleware(self) -> dict:
        """
        This method returns a dictionary of excluded middleware
        :return: exclude_middleware
        """
        return self.excluded_middleware

    def set_url(self, url) -> object:
        """
        This method sets the URL associated with this route
        :param url:
        :return: self / Route
        """
        self.url = url
        return self

    def get_url(self) -> str:
        """
        This method returns the URL for the route
        :return: url
        """
        return self.url

    def get_route(self) -> dict:
        """
        This method returns a dictionary representation of the route, this allows external
        code to pull the internal state of the route for routing purposes
        :return: route : dict
        """
        route = {}
        assert isinstance(self.get_name(), str), "Route name must be of type str"
        route["name"] = self.get_name()
        assert isinstance(self.get_verb(), int), "Route HTTP verbs must be defined using Sego.Routing.Verb.HTTP_<verb>"
        assert self.get_verb() in [Verb.HTTP_GET, \
                                   Verb.HTTP_POST, \
                                   Verb.HTTP_PUT, \
                                   Verb.HTTP_DELETE], "Route HTTP verbs must be defined using Sego.Routing.Verb.HTTP_<verb>"
        route["verb"] = self.get_verb()
        assert isinstance(self.get_controller(), str), "Route controller must be of type str"
        route["controller"] = self.get_controller()
        assert isinstance(self.get_action(), str), "Route action must be of type str"
        route["action"] = self.get_action()
        assert isinstance(self.get_url(), str), "Route url must be of type str"
        route["url"] = self.get_url()
        assert  isinstance(self.get_middleware(),dict)," Middleware must be of type dict"
        route["middleware"] = self.get_middleware()
        assert isinstance(self.get_excluded_middleware(),dict),"Excluded middleware must be stored in a dict"
        route["excluded_middleware"] = self.get_excluded_middleware()

        return route



from .SegoBaseException import SegoBaseException


class InternalServerErrorException500(SegoBaseException):
    message = '500 Internal Server Error'
    """ The server has
encountered a situation it doesn't know
how to handle. """


class NotImplementedException501(SegoBaseException):
    message = '501 Not Implemented'
    """ The request
method is not supported by the server
and cannot be handled. The only methods
that servers are required to support
(and therefore that must not return this
code) are <code>GET</code> and
<code>HEAD</code> . """


class BadGatewayException502(SegoBaseException):
    message = '502 Bad Gateway'
    """ This error
response means that the server, while
working as a gateway to get a response
needed to handle the request, got an
invalid response. """


class ServiceUnavailableException503(SegoBaseException):
    message = '503 Service Unavailable'
    """ The server is not
ready to handle the request. Common
causes are a server that is down for
maintenance or that is overloaded. Note
that together with this response, a
user-friendly page explaining the
problem should be sent. This responses
should be used for temporary conditions
and the <code>Retry-After:</code> HTTP
header should, if possible, contain the
estimated time before the recovery of
the service. The webmaster must also
take care about the caching-related
headers that are sent along with this
response, as these temporary condition
responses should usually not be cached.
"""


class GatewayTimeoutException504(SegoBaseException):
    message = '504 Gateway Timeout'
    """ This error
response is given when the server is
acting as a gateway and cannot get a
response in time. """


class HTTPVersionNotSupportedException505(SegoBaseException):
    message = '505 HTTP Version Not Supported'
    """ The HTTP version
used in the request is not supported by
the server. """


class VariantAlsoNegotiatesException506(SegoBaseException):
    message = '506 Variant Also Negotiates'
    """ The server has an
internal configuration error: the chosen
variant resource is configured to engage
in transparent content negotiation
itself, and is therefore not a proper
end point in the negotiation process.
"""


class InsufficientStorageException507(SegoBaseException):
    message = '507 Insufficient Storage'
    """ The method could
not be performed on the resource because
the server is unable to store the
representation needed to successfully
complete the request. """


class LoopDetectedException508(SegoBaseException):
    message = '508 Loop Detected'
    """ The server
detected an infinite loop while
processing the request. """


class NotExtendedException510(SegoBaseException):
    message = '510 Not Extended'
    """ Further
extensions to the request are required
for the server to fulfill it. """


class NetworkAuthenticationRequiredException511(SegoBaseException):
    message = '511 Network Authentication Required'
    """ The 511 status
code indicates that the client needs to
authenticate to gain network access. """



# ************************************************************************#
# Title:                    Sego  Exception                               #
# Description:              This class is the main Exception for Sego     #
# Author:                   Sambe Consulting <development@sambe.co.za>a     #
# Original Date:            29 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

class SegoBaseException(Exception):
    """Base class for other exceptions in Sego"""

    exception_name = "SegoBaseException"
    message = "Sego error"

    def __init__(self, message=None, exception_name=None):
        if exception_name is not None:
            self.exception_name = exception_name
        else:
            self.exception_name = self.__class__.__name__
        if message is not None:
            self.message = message



        super().__init__(self.message)

    def __str__(self):
        return f"""A Sego Exception has occurred\n
                     Exception name: {self.exception_name},
                     Exception message: {self.message}
                    """



from .SegoBaseException import SegoBaseException


class BadRequestException400(SegoBaseException):
    message = '400 Bad Request'
    """ The server could
not understand the request due to
invalid syntax. """


class UnauthorizedException401(SegoBaseException):
    message = '401 Unauthorized'
    """ Although the HTTP
standard specifies "unauthorized",
semantically this response means
"unauthenticated". That is, the client
must authenticate itself to get the
requested response. """


class PaymentRequiredException402(SegoBaseException):
    message = '402 Payment Required'
    """ This response
code is reserved for future use. The
initial aim for creating this code was
using it for digital payment systems,
however this status code is used very
rarely and no standard convention
exists. """


class ForbiddenException403(SegoBaseException):
    message = '403 Forbidden'
    """ The client does
not have access rights to the content;
that is, it is unauthorized, so the
server is refusing to give the requested
resource. Unlike 401, the client's
identity is known to the server. """


class NotFoundException404(SegoBaseException):
    message = '404 Not Found'
    """ The server can
not find the requested resource. In the
browser, this means the URL is not
recognized. In an API, this can also
mean that the endpoint is valid but the
resource itself does not exist. Servers
may also send this response instead of
403 to hide the existence of a resource
from an unauthorized client. This
response code is probably the most
famous one due to its frequent
occurrence on the web. """


class MethodNotAllowedException405(SegoBaseException):
    message = '405 Method Not Allowed'
    """ The request
method is known by the server but has
been disabled and cannot be used. For
example, an API may forbid DELETE-ing a
resource. The two mandatory methods,
<code>GET</code> and <code>HEAD</code> ,
must never be disabled and should not
return this error code. """


class NotAcceptableException406(SegoBaseException):
    message = '406 Not Acceptable'
    """ This response is
sent when the web server, after
performing <a href="/en-US/docs/Web/HTTP
/Content_negotiation#server-
driven_negotiation">server-driven
content negotiation</a> , doesn't find
any content that conforms to the
criteria given by the user agent. """


class ProxyAuthenticationRequiredException407(SegoBaseException):
    message = '407 Proxy Authentication Required'
    """ This is similar
to 401 but authentication is needed to
be done by a proxy. """


class RequestTimeoutException408(SegoBaseException):
    message = '408 Request Timeout'
    """ This response is
sent on an idle connection by some
servers, even without any previous
request by the client. It means that the
server would like to shut down this
unused connection. This response is used
much more since some browsers, like
Chrome, Firefox 27+, or IE9, use HTTP
pre-connection mechanisms to speed up
surfing. Also note that some servers
merely shut down the connection without
sending this message. """


class ConflictException409(SegoBaseException):
    message = '409 Conflict'
    """ This response is
sent when a request conflicts with the
current state of the server. """


class GoneException410(SegoBaseException):
    message = '410 Gone'
    """ This response is
sent when the requested content has been
permanently deleted from server, with no
forwarding address. Clients are expected
to remove their caches and links to the
resource. The HTTP specification intends
this status code to be used for
"limited-time, promotional services".
APIs should not feel compelled to
indicate resources that have been
deleted with this status code. """


class LengthRequiredException411(SegoBaseException):
    message = '411 Length Required'
    """ Server rejected
the request because the <code>Content-
Length</code> header field is not
defined and the server requires it. """


class PreconditionFailedException412(SegoBaseException):
    message = '412 Precondition Failed'
    """ The client has
indicated preconditions in its headers
which the server does not meet. """


class PayloadTooLargeException413(SegoBaseException):
    message = '413 Payload Too Large'
    """ Request entity is
larger than limits defined by server;
the server might close the connection or
return an <code>Retry-After</code>
header field. """


class URITooLongException414(SegoBaseException):
    message = '414 URI Too Long'
    """ The URI requested
by the client is longer than the server
is willing to interpret. """


class UnsupportedMediaTypeException415(SegoBaseException):
    message = '415 Unsupported Media Type'
    """ The media format
of the requested data is not supported
by the server, so the server is
rejecting the request. """


class RangeNotSatisfiableException416(SegoBaseException):
    message = '416 Range Not Satisfiable'
    """ The range
specified by the <code>Range</code>
header field in the request can't be
fulfilled; it's possible that the range
is outside the size of the target URI's
data. """


class ExpectationFailedException417(SegoBaseException):
    message = '417 Expectation Failed'
    """ This response
code means the expectation indicated by
the <code>Expect</code> request header
field can't be met by the server. """


class ImateapotException418(SegoBaseException):
    message = '418 Im a teapot'
    """ The server
refuses the attempt to brew coffee with
a teapot. """


class MisdirectedRequestException421(SegoBaseException):
    message = '421 Misdirected Request'
    """ The request was
directed at a server that is not able to
produce a response. This can be sent by
a server that is not configured to
produce responses for the combination of
scheme and authority that are included
in the request URI. """


class UnprocessableEntityException422(SegoBaseException):
    message = '422 Unprocessable Entity'
    """ The request was
well-formed but was unable to be
followed due to semantic errors. """


class LockedException423(SegoBaseException):
    message = '423 Locked'
    """ The resource that
is being accessed is locked. """


class FailedDependencyException424(SegoBaseException):
    message = '424 Failed Dependency'
    """ The request
failed due to failure of a previous
request. """


class TooEarlyException425(SegoBaseException):
    message = '425 Too Early'
    """ Indicates that
the server is unwilling to risk
processing a request that might be
replayed. """


class UpgradeRequiredException426(SegoBaseException):
    message = '426 Upgrade Required'
    """ The server
refuses to perform the request using the
current protocol but might be willing to
do so after the client upgrades to a
different protocol. The server sends an
<a href="/en-US/docs/Web/HTTP/Headers/Up
grade"><code>Upgrade</code></a> header
in a 426 response to indicate the
required protocol(s). """


class PreconditionRequiredException428(SegoBaseException):
    message = '428 Precondition Required'
    """ The origin server
requires the request to be conditional.
This response is intended to prevent the
'lost update' problem, where a client
GETs a resource's state, modifies it,
and PUTs it back to the server, when
meanwhile a third party has modified the
state on the server, leading to a
conflict. """


class TooManyRequestsException429(SegoBaseException):
    message = '429 Too Many Requests'
    """ The user has sent
too many requests in a given amount of
time ("rate limiting"). """


class RequestHeaderFieldsTooLargeException431(SegoBaseException):
    message = '431 Request Header Fields Too Large'
    """ The server is
unwilling to process the request because
its header fields are too large. The
request may be resubmitted after
reducing the size of the request header
fields. """


class UnavailableForLegalReasonsException451(SegoBaseException):
    message = '451 Unavailable For Legal Reasons'
    """ The user-agent
requested a resource that cannot legally
be provided, such as a web page censored
by a government. """



from .SegoBaseException import SegoBaseException


class OKException200(SegoBaseException):
    message = '200 OK'
    """ The request has
succeeded. The meaning of the success
depends on the HTTP method: <ul>
<li><code>GET</code>: The resource has
been fetched and is transmitted in the
message body.</li>
<li><code>HEAD</code>: The entity
headers are in the message body.</li>
<li><code>PUT</code> or
<code>POST</code>: The resource
describing the result of the action is
transmitted in the message body.</li>
<li><code>TRACE</code>: The message body
contains the request message as received
by the server.</li> </ul> """


class CreatedException201(SegoBaseException):
    message = '201 Created'
    """ The request has
succeeded and a new resource has been
created as a result. This is typically
the response sent after
<code>POST</code> requests, or some
<code>PUT</code> requests. """


class AcceptedException202(SegoBaseException):
    message = '202 Accepted'
    """ The request has
been received but not yet acted upon. It
is noncommittal, since there is no way
in HTTP to later send an asynchronous
response indicating the outcome of the
request. It is intended for cases where
another process or server handles the
request, or for batch processing. """


class NonAuthoritativeInformationException203(SegoBaseException):
    message = '203 Non-Authoritative Information'
    """ This response
code means the returned meta-information
is not exactly the same as is available
from the origin server, but is collected
from a local or a third-party copy. This
is mostly used for mirrors or backups of
another resource. Except for that
specific case, the "200 OK" response is
preferred to this status. """


class NoContentException204(SegoBaseException):
    message = '204 No Content'
    """ There is no
content to send for this request, but
the headers may be useful. The user-
agent may update its cached headers for
this resource with the new ones. """


class ResetContentException205(SegoBaseException):
    message = '205 Reset Content'
    """ Tells the user-
agent to reset the document which sent
this request. """


class PartialContentException206(SegoBaseException):
    message = '206 Partial Content'
    """ This response
code is used when the <a href="/en-US/do
cs/Web/HTTP/Headers/Range"><code>Range</
code></a> header is sent from the client
to request only part of a resource. """


class MultiStatusException207(SegoBaseException):
    message = '207 Multi-Status'
    """ Conveys
information about multiple resources,
for situations where multiple status
codes might be appropriate. """


class AlreadyReportedException208(SegoBaseException):
    message = '208 Already Reported'
    """ Used inside a
<code>&lt;dav:propstat&gt;</code>
response element to avoid repeatedly
enumerating the internal members of
multiple bindings to the same
collection. """


class IMUsedException226(SegoBaseException):
    message = '226 IM Used'
    """ The server has
fulfilled a <code>GET</code> request for
the resource, and the response is a
representation of the result of one or
more instance-manipulations applied to
the current instance. """



# ************************************************************************#
# Title:                    Middleware API                                #
# Description:              This class is the main interface into  Sego's #
#                           middleware system                             #
# Author:                   Sambe Consulting <development@sambe.co.za>    #
# Original Date:            10 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from webob import Request, Response


class Middleware:
    PREPROCESS = 1
    POSTPROCESS = 2

    def process_request(self, request: Request):
        """
            This method is meant for processing requests in the middleware layer
            :param request:
            :return: request : webob.Request
        """
        pass

    def process_response(self,  request: Request, response: Response):
       """
        This method is meant for processing responses in the middleware layer
        :param request:
        :param response:
        :return:
       """
       pass



from .SegoBaseException import SegoBaseException
from .RoutingExceptions import *
from .ConfigurationException import *
from .HTTPVerbException import *
from .ClientExceptions import *
from .ClientErrorResponses import *


# ************************************************************************#
# Title:                    Sego Router                                   #
# Description:              This class handles all routing tasks for sego #
# Author:                   Kabelo Masemola <kn3rdydad@gmail.com>         #
# Original Date:            10 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from .Route import Route
from .Verb import Verb
from ..Exceptions import *
from parse import parse
from singleton_decorator import singleton
from confo.Confo import Confo
import ctypes


@singleton
class Router:
    def __init__(self):
        """
        This constructor sets up the Router object by instantiating routes,urls and config_manager
        """
        self.routes = []
        self.urls = []
        self.configuration_manager = Confo()

    def add_route(self, route: Route):
        """
        This method registers a new Route object
        :param route:
        :return:
        """
        route_parameter = route.get_route()
        self.validate(route_parameters=route_parameter)
        self.routes.append(route_parameter)
        self.urls.append(route_parameter["url"].rstrip("/"))

    def validate(self, route_parameters: dict):
        """
        This method is used to validate if a Route object has been given all the correct
        parameters and to enforce uniqueness constraints
        :param route_parameters:
        :return:
        """
        message = "Route with URL: " + route_parameters["url"] + " already exists."
        if route_parameters["url"].rstrip("/") in self.urls: raise RouteAlreadyExistsException(message)
        if route_parameters["name"] in list([r["name"] for r in self.routes]): raise UniqueNameException(
            " Use unique route names, " + route_parameters["name"] + " already exists.")

    def find_route(self, request):
        """
        This method to search for a route that matches the request
        :param request:
        :return:
        """

        def method_validator(request_method, route_method):
            """
            This method is used to validate if HTTP verbs match
            :param request_method:
            :param route_method:
            :return:
            """
            if Verb().validate(request_method) == route_method:
                return True
            else:
                return False

        request_path = request.path
        if request_path.strip().lower() != "/":
            request_path = request_path.rstrip("/")

        for route in self.routes:
            parse_result = parse(route["url"], request_path)
            if parse_result is not None:
                return (route, parse_result.named) if method_validator(request.method, route["verb"]) else (None, None)
            else:
                parse_result = parse(route["url"].rstrip("/"), request_path.rstrip("/"))
                if parse_result is not None:
                    return (route, parse_result.named) if method_validator(request.method, route["verb"]) else (
                    None, None)

        return None, None

    def register_routes(self, route_paths):
        """
        This method loads all defined routes by registering the
        relevant modules
        :param route_paths:
        :return:
        """
        self.route_definitions = __import__(route_paths)

    def dynamic_import(self, controller):
        """
        This method is used to dynamically load the relevant controller and action into
        a variable
        :param controller:
        :return: _class : Object
        """
        controller_components = controller.split(".")
        controller_components.insert(0, "Controllers")
        controller_components.insert(0, "app")
        controller_path = '.'.join(controller_components)
        module = __import__(name=controller_path, fromlist=controller_components)
        _class = getattr(module, controller_components[-1])
        return _class

    def build_handler(self, route_parameters):
        """
        This special method is used when using the Flask style routes, it using meta-programming to
        create a controller for the method then links the action, them makes them callable ,the send control
        back to the normal execution path
        :param route_parameters:
        :return:
        """
        Controller = route_parameters["controller"]
        action = route_parameters["action"]
        if Controller == "SegoBaseController":
            handler_id = self.configuration_manager.get("dynamic_routes", action)
            handler = ctypes.cast(handler_id, ctypes.py_object).value
            return handler

        _class = self.dynamic_import(Controller)
        return getattr(_class(), action)

    def get_routes(self):
        """
        This method returns all load routes for debug purposes
        :return: routes
        """
        return self.routes



import pathlib
from setuptools import setup, find_packages
import os
import fnmatch
import sysconfig
from setuptools.command.build_py import build_py as _build_py
EXCLUDE_FILES = [
    "sego/tests/*"
]


# The directory containing this file
HERE = os.path.dirname(os.path.realpath(__file__))

# The text of the README file
with open(HERE + "/README.md", encoding='utf-8') as f:
    README = f.read()
with open(HERE + '/requirements.txt') as f:
    required = f.read().splitlines()
    setup(
        name="sego",
        version="0.1.0.8",
        description="A web and micro-service framework for data analysts",
        long_description_content_type="text/markdown",
        long_description=README,
        url="https://github.com/sambe-consulting/sego",
        author="Sambe Consulting",
        author_email="development@sambe.co.za",
        license="Apache License 2.0",
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
        ],
        packages=find_packages(exclude=("sego/tests",)),
        include_package_data=True,
        install_requires=required,

    )
# This call to setup() does all the work



from . import SegoBaseException


class BadRequestException400(SegoBaseException):
    message = "400 Bad Request"


class UnauthorizedException402(SegoBaseException):
    message = "401 Unauthorized"


class ForbiddenException403(SegoBaseException):
    message = "403 Forbidden"


class NotFoundException404(SegoBaseException):
    message = "404 Not Found"


class MethodNotAllowedException405(SegoBaseException):
    message = "405 Method Not Allowed"


class NotAcceptableException406(SegoBaseException):
    message = "406 Not Acceptable"

class ProxyAuthenticationRequiredException407(SegoBaseException):
    message = "407 Proxy Authentication Required"


class ContinueException100:

    message = '100 Continue'
    """ This interim response indicates
    that everything so far is OK and that
    the client should continue the
    request, or ignore the response if
    the request is already finished. """






from .SegoBaseException import SegoBaseException


class MultipleChoiceException300(SegoBaseException):
    message = '300 Multiple Choice'
    """ The request has
more than one possible response. The
user-agent or user should choose one of
them. (There is no standardized way of
choosing one of the responses, but HTML
links to the possibilities are
recommended so the user can pick.) """


class MovedPermanentlyException301(SegoBaseException):
    message = '301 Moved Permanently'
    """ The URL of the
requested resource has been changed
permanently. The new URL is given in the
response. """


class FoundException302(SegoBaseException):
    message = '302 Found'
    """ This response
code means that the URI of requested
resource has been changed
<em>temporarily</em> . Further changes
in the URI might be made in the future.
Therefore, this same URI should be used
by the client in future requests. """


class SeeOtherException303(SegoBaseException):
    message = '303 See Other'
    """ The server sent
this response to direct the client to
get the requested resource at another
URI with a GET request. """


class NotModifiedException304(SegoBaseException):
    message = '304 Not Modified'
    """ This is used for
caching purposes. It tells the client
that the response has not been modified,
so the client can continue to use the
same cached version of the response. """


class UseProxyException305(SegoBaseException):
    message = '305 Use Proxy'
    """ Defined in a
previous version of the HTTP
specification to indicate that a
requested response must be accessed by a
proxy. It has been deprecated due to
security concerns regarding in-band
configuration of a proxy. """


class unusedException306(SegoBaseException):
    message = '306 unused'
    """ This response
code is no longer used; it is just
reserved. It was used in a previous
version of the HTTP/1.1 specification.
"""


class TemporaryRedirectException307(SegoBaseException):
    message = '307 Temporary Redirect'
    """ The server sends
this response to direct the client to
get the requested resource at another
URI with same method that was used in
the prior request. This has the same
semantics as the <code>302 Found</code>
HTTP response code, with the exception
that the user agent <em>must not</em>
change the HTTP method used: If a
<code>POST</code> was used in the first
request, a <code>POST</code> must be
used in the second request. """


class PermanentRedirectException308(SegoBaseException):
    message = '308 Permanent Redirect'
    """ This means that
the resource is now permanently located
at another URI, specified by the
<code>Location:</code> HTTP Response
header. This has the same semantics as
the <code>301 Moved Permanently</code>
HTTP response code, with the exception
that the user agent <em>must not</em>
change the HTTP method used: If a
<code>POST</code> was used in the first
request, a <code>POST</code> must be
used in the second request. """



from .SegoBaseException import SegoBaseException


class ContinueException100(SegoBaseException):
    message = '100 Continue'
    """ This interim
response indicates that everything so
far is OK and that the client should
continue the request, or ignore the
response if the request is already
finished. """


class SwitchingProtocolException101(SegoBaseException):
    message = '101 Switching Protocol'
    """ This code is sent
in response to an <a href="/en-US/docs/W
eb/HTTP/Headers/Upgrade"><code>Upgrade</
code></a> request header from the
client, and indicates the protocol the
server is switching to. """


class ProcessingException102(SegoBaseException):
    message = '102 Processing'
    """ This code
indicates that the server has received
and is processing the request, but no
response is available yet. """


class EarlyHintsException103(SegoBaseException):
    message = '103 Early Hints'
    """ This status code
is primarily intended to be used with
the <a href="/en-US/docs/Web/HTTP/Header
s/Link"><code>Link</code></a> header,
letting the user agent start <a
href="/en-US/docs/Web/HTML/Preloading_co
ntent">preloading</a> resources while
the server prepares a response. """



# ************************************************************************#
# Title:                    Views  API                                    #
# Description:              This class handles the rendering of data      #
# Author:                   Sambe Consulting <development@sambe.co.za>    #
# Original Date:            20 March 2021                                 #
# Version:                  0.1.0                                         #
# ************************************************************************#

from jinja2 import Environment, FileSystemLoader
from singleton_decorator import singleton
import os


@singleton
class Views(object):
    def __init__(self, view_path):
        """
        The Constructor takes in a path where jinja templates are stored, then uses it to
        create a jinja templating environment
        :param view_path:
        """
        self.view_environment = Environment(
            loader=FileSystemLoader(os.path.abspath(view_path))
        )

    def render_view(self, view_name, context=None):
        """
        This method is used at the controller layer to render views
        :param view_name:
        :param context:
        :return:
        """
        if context is None:
            context = {}
        return self.view_environment.get_template(view_name).render(**context)
